/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CFile***********************************************************************

  FileName    [cutptl.c]

  PackageName [BDS]

  Synopsis    [Cut PTL gates from final synthesized circuit]

  Description []

  SeeAlso     []

  Author      [Anda Congguang Yang]

  Copyright   [The file was created in Department of Electrical & Computer
	Engineering, University of Massachusetts Amherst. All source codes
	in BDS package are CONFIDENTIAL before a formal release is
	announced. The author reserve all right of the package.]

******************************************************************************/

#include "lopt.h"
#include <stdio.h>

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int bdsCutPtlFromGeneralNetlistRecursive ARGS((FactorTreeNode*,st_table*,st_table*,st_table*));
static int BDS_PrintBlifFromFtreeNodeList ARGS((BnetNetwork*,bdsOptions*,st_table*,st_table*,st_table*,char**));
static FactorTreeNode* bdsCreateNewPseudoPrimary ARGS((int));
static int bdsPrintBlifFromFtreeNodeList ARGS((st_table*,st_table*,char**,FILE*));
static int bdsPrintOneNodePrintVarNames ARGS((FILE*,FactorTreeNode*,FactorTreeNode*,char**,st_table*));
static int bdsIfDuplicate ARGS((lsList,FactorTreeNode*,char**));
static int bdsPrintOneNodeCutPtl ARGS((FILE*,FactorTreeNode*,FactorTreeNode*,char**,st_table*));
static int bdsPrintOneNodePrintFunction ARGS((FILE*,FactorTreeNode*));

/**Function********************************************************************

  Synopsis    [Cut PTL gates from general netlist]

  Description [Since SIS mapper is weak at handling PTL and XOR mapping, those
  gates are cut from the final synthesis netlist. Submit the AND/OR only part
  to SIS mapper. After mapping, two netlist are concateted togather. Return 1
  if successful; o otherwise.]

  SideEffects [Factoring trees are modified]

  SeeAlso     []

  LastDate    [3/1/99]

******************************************************************************/
extern
int
BDS_CutPtlFromGeneralNetlist(
  BnetNetwork *net,
  FactorTreeNode **FTreeNode,
  char **PIvariables,
  st_table *sharingFnodeTbl,
  bdsOptions *option)
{
  int i, result;
  FactorTreeNode **q, *dummyPO;
  BnetNode *POnode;
  st_table *fnode2pseudoOutput, *fnode2pseudoInput;

  /* Initialize fnode->pseudo primary output table */
  fnode2pseudoOutput = st_init_table(st_ptrcmp,st_ptrhash);

  /* Initialize fnode->pseudo primary input table */
  fnode2pseudoInput = st_init_table(st_ptrcmp,st_ptrhash);

  for (i = 0; i < net->npos; i++) {
      result = bdsCutPtlFromGeneralNetlistRecursive(FTreeNode[i],fnode2pseudoInput,fnode2pseudoOutput,sharingFnodeTbl);
      if (result == 0) return(0);

      /* Add original POs. Their names are copied to new fnodes */
      if (Fnode_Operator(FTreeNode[i]) == BDS_BDD_AND || Fnode_Operator(FTreeNode[i]) == BDS_BDD_OR) {
	  dummyPO = ALLOC(FactorTreeNode, 1);
	  dummyPO->value = BDS_FTREE_PSEUDO;
	  dummyPO->name = ALLOC(char, 20);
	  st_lookup(net->hash,net->outputs[i],(char **) &POnode);
	  strcpy(dummyPO->name, POnode->name);
          if (st_add_direct(fnode2pseudoOutput, (char *) FTreeNode[i], (char *) dummyPO) == ST_OUT_OF_MEM) {
	      printf("Out of memory !");
	      return(0);
          }
      }
  }

  /* Print out the modified netlist in BLIF format */
  result = BDS_PrintBlifFromFtreeNodeList(net,option,fnode2pseudoInput,fnode2pseudoOutput,sharingFnodeTbl,PIvariables);
  if (result == 0) return(0);

  st_free_table(fnode2pseudoOutput);
  st_free_table(fnode2pseudoInput);

} /* end of BDS_CutPtlFromGeneralNetlist */

/**Function********************************************************************

  Synopsis    [Recursive program for BDS_CutPtlFromGeneralNetlist]

  Description [Return 1 if successful; o otherwise.]

  SideEffects []

  SeeAlso     []

  LastDate    [3/1/99]

******************************************************************************/
static 
int 
bdsCutPtlFromGeneralNetlistRecursive(
  FactorTreeNode *fnode,
  st_table *fnode2pseudoInput,
  st_table *fnode2pseudoOutput,
  st_table *sharingFnodeTbl)
{
  int result;
  FactorTreeNode *top, *sib, *ctl, *pseudoPI, *dummyPO, *listItem;
  lsList fanins;
  lsGen siblingList;
  lsHandle sibHandle;
  lsStatus status;

  /* If terminal, go back. */
  if (!Fnode_Internal(fnode)) return(1);

  top = Fnode_Regular(fnode);

  /* If a pseudo input has been assigned to this fnode, go back
  */
  if (st_is_member(fnode2pseudoInput, (char *) top)) return(1);

#ifdef PTL
  if (Fnode_Operator(top) == BDS_BDD_MUX) {
      ctl = top->control;
  }
#endif

  /* Traverse to the bottom */
  fanins = top->fanins;
  siblingList = lsStart(fanins);

  while (lsNext(siblingList, (lsGeneric *)&sib, NULL) == LS_OK) {
      result = bdsCutPtlFromGeneralNetlistRecursive(sib,fnode2pseudoInput,fnode2pseudoOutput,sharingFnodeTbl);
      if (result == 0) return(0);
  }
  lsFinish(siblingList);

#ifdef PTL
  if (Fnode_Operator(top) == BDS_BDD_MUX) {
      result = bdsCutPtlFromGeneralNetlistRecursive(ctl,fnode2pseudoInput,fnode2pseudoOutput,sharingFnodeTbl);
      if (result == 0) return(0);
  }
#endif

  /* If the operator of fnode is one of xor, xnor and mux, create pseudo primary
     outputs for all its siblings. Results are stored in fnode2pseudoOutput.
     A Dummy fnode is also created to hold its name.
     If one sibling is of the same operator, don't do anything.
  */
  if (Fnode_Operator(fnode) == BDS_BDD_XOR  || 
      Fnode_Operator(fnode) == BDS_BDD_XNOR || Fnode_Operator(fnode) == BDS_BDD_MUX) {

      siblingList = lsStart(fanins);
      while (lsNext(siblingList, (lsGeneric *)&sib, NULL) == LS_OK) {
	  if ((Fnode_Operator(sib) == BDS_BDD_AND || Fnode_Operator(sib) == BDS_BDD_OR) && Fnode_Internal(sib)) {
              if (!st_is_member(fnode2pseudoOutput, (char *) Fnode_Regular(sib))) {
		  dummyPO = bdsCreateNewPseudoPrimary(1);
	          if (st_add_direct(fnode2pseudoOutput, (char *) sib, (char *) dummyPO) == ST_OUT_OF_MEM) {
	              printf("Out of memory !");
		      return(0);
                  }
              }
	  }
      }
      lsFinish(siblingList);

#ifdef PTL
      if (Fnode_Operator(top) == BDS_BDD_MUX) {
          if (Fnode_Operator(ctl) == BDS_BDD_AND || Fnode_Operator(ctl) == BDS_BDD_OR) {
              if (!st_is_member(fnode2pseudoOutput, (char *) Fnode_Regular(ctl))) {
		  dummyPO = bdsCreateNewPseudoPrimary(1);
	          if (st_add_direct(fnode2pseudoOutput, (char *) ctl, (char *) dummyPO) == ST_OUT_OF_MEM) {
	              printf("Out of memory !");
		      return(0);
                  }
              }
          }
      }
#endif

  }
  else { 
      /* Operator = BDS_BDD_AND or BDS_BDD_OR. Create pseudo PIs. A new primary input fnode
      ** is created to hold new PI name.
      */
      siblingList = lsStart(fanins);
      while (lsNext(siblingList, (lsGeneric *)&sib, (lsHandle *)&sibHandle) == LS_OK) {
	  if (!Fnode_Internal(sib)) { /* PI */
	      if (!st_is_member(fnode2pseudoInput, (char *) Fnode_Regular(sib))) {	
                  if (st_add_direct(fnode2pseudoInput, (char *) Fnode_Regular(sib), NULL) == ST_OUT_OF_MEM) {
	              printf("Out of memory !");
	              return(0);
                  }
	      }
	  }
	  else if ((Fnode_Operator(sib) == BDS_BDD_XOR || Fnode_Operator(sib) == BDS_BDD_XNOR ||
				                Fnode_Operator(sib) == BDS_BDD_MUX) && Fnode_Internal(sib)) {
	      if (!st_lookup(fnode2pseudoInput, (char *) Fnode_Regular(sib), (char **) &pseudoPI)) {

                  pseudoPI = bdsCreateNewPseudoPrimary(0);
                  if (st_add_direct(fnode2pseudoInput, (char *) pseudoPI, NULL) == ST_OUT_OF_MEM) {
	              printf("Out of memory !");
	              return(0);
                  }
              }
	      /* Remove sib from siblist, patch new PI */
	      if (Fnode_IsComplement(sib)) {
	          status = lsInAfter(siblingList, (lsGeneric) Fnode_Complement(pseudoPI), NULL);
	      }
	      else {
		  status = lsInAfter(siblingList, (lsGeneric) pseudoPI, NULL);
	      }
	      status = lsDelBefore(siblingList, (lsGeneric*) &listItem);
	      status = lsNext(siblingList, (lsGeneric *)&sib, (lsHandle *)&sibHandle);
          }
      }
      lsFinish(siblingList);
  }

  return(1);

} /* end of bdsCutPtlFromGeneralNetlistRecursive */

/**Function********************************************************************

  Synopsis    [Create an fnode to hold pseudo POs' and PIs' names]

  Description []

  SideEffects []

  SeeAlso     []

  LastDate    [3/2/99]

******************************************************************************/
static 
FactorTreeNode *
bdsCreateNewPseudoPrimary(
  int inORout)
{
  char *pseudoIn, *pseudoOut, *pseudo_name;
  FactorTreeNode *ret_fnode;
  static int pi_count, po_count;

  pseudo_name = ALLOC(char, 10);
  ret_fnode = ALLOC(FactorTreeNode, 1);

  if (inORout) {
      sprintf(pseudo_name, "psuPO%d", po_count);
      po_count++;
  }
  else {
      sprintf(pseudo_name, "psuPI%d", pi_count);
      pi_count++;
  }

  ret_fnode->value = BDS_FTREE_PSEUDO;
  ret_fnode->name = pseudo_name;
  ret_fnode->polarity = 1;

  return(ret_fnode);

} /* end of bdsCreateNewPseudoPrimary */

/**Function********************************************************************

  Synopsis    []

  Description [Return 1 if successful; o otherwise.]

  SideEffects []

  SeeAlso     []

  LastDate    [3/1/99]

******************************************************************************/
static 
int
BDS_PrintBlifFromFtreeNodeList(
  BnetNetwork *net,
  bdsOptions *option,
  st_table *fnode2pseudoInput,
  st_table *fnode2pseudoOutput,
  st_table *sharingFnodeTbl,
  char **PIvariables)
{
  FILE *dfp = NULL;
  int result;
  char *dumpfile, *dumpfile_name, *ckt_name;


  /* Generate file_name as cktname.cutptl.blif */
  ckt_name = BDS_GetFileHead(option->file);
  dumpfile_name = ALLOC(char, strlen(ckt_name)+32); /* file_name should not longer than 100 chars but can be so JS*/
  strcpy(dumpfile_name, ckt_name);
  strcat(dumpfile_name, ".cutptl.blif");

  /* Put this file in the run dir */
  dumpfile = BDS_CreateFname(option, dumpfile_name);

  dfp = fopen(dumpfile,"w");
  if (dfp == NULL) fail("Can not open file to write !");

  result = bdsPrintBlifFromFtreeNodeList(fnode2pseudoInput,fnode2pseudoOutput,PIvariables,dfp);
  if (result == 0) return(0);

  if (fclose(dfp) == EOF) return(0);
  FREE(dumpfile); FREE(dumpfile_name); FREE(ckt_name);

  return(1);

} /* end of BDS_PrintBlifFromFtreeNodeList */

/**Function********************************************************************

  Synopsis    []

  Description [Return 1 if successful; o otherwise.]

  SideEffects []

  SeeAlso     []

  LastDate    [3/1/99]

******************************************************************************/
static 
int
bdsPrintBlifFromFtreeNodeList(
  st_table *fnode2pseudoInput,
  st_table *fnode2pseudoOutput,
  char **PIvariables,
  FILE *fp)
{
  FactorTreeNode *fnode, *dummy_node;
  st_table       *visited;
  st_generator   *gen = NULL;
  int            result;
  int            i, j;
  BnetNode       *POnode;
  FactorTreeNode *top, *dp, *dm, *ctl;
  lsList	 newPIs;

  /* Initialize symbol table for visited nodes. */
  visited = st_init_table(st_ptrcmp, st_ptrhash);
  if (visited == NULL) return(0);

  /* Collect all the nodes on the ftrees */
  gen = st_init_gen(fnode2pseudoOutput);
  if (gen == NULL) return(0);
  while (st_gen(gen, (char **) &fnode, NULL)) {
      result = bdsCollectFNodeslsList(visited,fnode);
      if (result == 0) return(0);
  }
  st_free_gen(gen);

  result = fprintf(fp,"# The file is generated by %s\n",BDS_VERSION);
  if (result == EOF) return(0);
  result = fprintf(fp,"#\n");
  if (result == EOF) return(0);

  /* Write header information */
  result = fprintf(fp, ".model cutptl\n");
  if (result == EOF) return(0);

  /* Write the inputs & outputs list by scanning the list in blif file 
     Since fnode2pseudoInput may contains multi-definition of a PI, filter
     it before print .inputs 
  */
  result = fprintf(fp, ".inputs ");
  if (result == EOF) return(0);

  newPIs = lsCreate();
  gen = st_init_gen(fnode2pseudoInput);
  if (gen == NULL) return(0);
  while (st_gen(gen, (char **) &fnode, NULL)) {
      if (bdsIfDuplicate(newPIs, fnode, PIvariables)) {
	  st_delete(fnode2pseudoInput, (char **) &fnode, NULL);
      }
      else {
	  lsNewEnd(newPIs, (lsGeneric) fnode, NULL);
          if (Fnode_Regular(fnode)->value == BDS_FTREE_PSEUDO) {
              result = fprintf(fp, "%s ", Fnode_Regular(fnode)->name);
              if (result == EOF) return(0);
          }
          else {
              result = fprintf(fp, "%s ", PIvariables[Fnode_Regular(fnode)->value]);
              if (result == EOF) return(0);
          }
      }
  }
  st_free_gen(gen);
  lsDestroy(newPIs, NULL);
  result = fprintf(fp,"\n");
  if (result == EOF) return(0);

  result = fprintf(fp,".outputs ");
  if (result == EOF) return(0);
  gen = st_init_gen(fnode2pseudoOutput);
  if (gen == NULL) return(0);
  while (st_gen(gen, (char **) &fnode, (char **)&dummy_node)) {
      result = fprintf(fp, "%s ", Fnode_Regular(dummy_node)->name);
      if (result == EOF) return(0);
  }
  st_free_gen(gen);
  result = fprintf(fp, "\n");
  if (result == EOF) return(0);

  /* Write outputs -> ftree first, this will put pos in the beginning of BLIF file */
  gen = st_init_gen(fnode2pseudoOutput);
  if (gen == NULL) return(0);
  while (st_gen(gen, (char **) &fnode, (char **)&dummy_node)) {
      result = bdsPrintOneNodeCutPtl(fp, fnode, dummy_node, PIvariables, fnode2pseudoOutput);
      if (result == 0) return(0);
  }
  st_free_gen(gen);

  /* Write all other internal nodes */
  gen = st_init_gen(visited);
  if (gen == NULL) return(0);
  while (st_gen(gen, (char **) &fnode, NULL)) {

      /* Skip POs, they have been dumped */
      if (st_is_member(fnode2pseudoOutput, (char *) fnode) == 1) { /* PO node */
          continue;
      }
      /* Skip terminal nodes */
      if (Fnode_Regular(fnode)->value != BDS_FTREE_INTERNAL) {
          continue;
      }

      result = bdsPrintOneNodeCutPtl(fp, fnode, NULL, PIvariables, fnode2pseudoOutput);
      if (result == 0) return(0);
  }

  /* write end */
  result = fprintf(fp, ".end\n");
  if (result == EOF) return(0);

  st_free_gen(gen);
  st_free_table(visited);

  return(1);

} /* end of bdsPrintBlifFromFtreeNodeList */

/**Function********************************************************************

  Synopsis    []

  Description [Return 1 if successful; o otherwise.]

  SideEffects []

  SeeAlso     []

  LastDate    [3/2/99]

******************************************************************************/
static 
int
bdsPrintOneNodeCutPtl(
  FILE *fp, 
  FactorTreeNode *fnode, 
  FactorTreeNode *dummy_node, 
  char **PIvariables, 
  st_table *fnode2pseudoOutput)
{
  int result;
  FactorTreeNode *top;

  top = Fnode_Regular(fnode);

  /* Screen out the possibility that po node is a terminal */
  if (top->value != BDS_FTREE_INTERNAL && dummy_node != NULL) {
      result = fprintf(fp,".names ");
      if (result == EOF) return(0);
      result = fprintf(fp,"%s %s\n",PIvariables[top->value],dummy_node->name);
      if (result == EOF) return(0);
      result = fprintf(fp,"%c 1\n",(top->polarity)?'0':'1');
      if (result == EOF) return(0);

      return(1);
  }

  /* print variable names */
  result = bdsPrintOneNodePrintVarNames(fp,top,dummy_node,PIvariables,fnode2pseudoOutput);
  if (result == 0) return(0);

  /* print functionality */
  result = bdsPrintOneNodePrintFunction(fp,top);

  return(result);

} /* end of bdsPrintOneNodeCutPtl */

/**Function********************************************************************

  Synopsis    [Print out the support of this fnode]

  Description [In cases like, (a+b)' and (fg+f'h)', instead of printing a'b' and
  fg'+f'h', inverters are inserted to keep their original forms. Dummy variables
  are introduced by alloc new FactorTreeNode. The memory allocated for them are
  leaked. Return 1 if successful, 0 otherwise]

  SideEffects [None]

  SeeAlso     []

  Last date   [2/28/99]

*****************************************************************************/
static
int
bdsPrintOneNodePrintVarNames(
  FILE *fp,
  FactorTreeNode *top,
  FactorTreeNode *dummy_node,
  char **PIvariables,
  st_table *fnode2pseudoOutput)
{
  int result;
  FactorTreeNode *sib, *ctl, *inverter;
  FactorTreeNode *po;
  lsList fanins;
  lsGen fanin_list;

  fanins = Fnode_Fanins(top);

  /* In case of (a+b)' and (fg+f'h)', insert an inverter */
  if (Fnode_Operator(top) == BDS_BDD_OR && Fnode_Negative(top) ||
               Fnode_Operator(top) == BDS_BDD_MUX && Fnode_Negative(top)) {
      inverter = ALLOC(FactorTreeNode, 1);

      result = fprintf(fp, ".names ");
      if (result == EOF) return(0);

      /* Print intermediate name */
      result = fprintf(fp,"%p ",inverter);
      if (result == EOF) return(0);

      /* Print functional node's name */
      if (dummy_node == NULL) { /* internal node */
          result = fprintf(fp,"%p\n",top);
          if (result == EOF) return(0);
      }
      else { /* PO node, print its name */
          result = fprintf(fp,"%s\n",dummy_node->name);
          if (result == EOF) return(0);
      }

      /* Print inverter's functionality */
      result = fprintf(fp,"0 1\n");
      if (result == EOF) return(0);
  }

  /* Print names under normal conditions */
  result = fprintf(fp, ".names ");
  if (result == EOF) return(0);

#ifdef PTL
  /* Print name corresponding to f in MUX */
  if (top->operator == BDS_BDD_MUX) { /* a MUX node */
      ctl = top->control;
      if (Fnode_Regular(ctl)->value == BDS_FTREE_INTERNAL) { /* internal node */
          if (!st_lookup(fnode2pseudoOutput, (char *) Fnode_Regular(ctl), (char **) &po)) {
              result = fprintf(fp,"%p ",Fnode_Regular(ctl));
              if (result == EOF) return(0);
          }
          else {
              result = fprintf(fp,"%s ",po->name);
              if (result == EOF) return(0);
          }
      }
      else { /* terminal node, print pi name */
          result = fprintf(fp,"%s ",PIvariables[Fnode_Regular(ctl)->value]);
          if (result == EOF) return(0);
      }
  }
#endif

  /* Print support names on this fnode */
  fanin_list = lsStart(fanins);
  while (lsNext(fanin_list, (lsGeneric *)&sib, NULL) != LS_NOMORE) {

      if (Fnode_Regular(sib)->value == BDS_FTREE_INTERNAL) { /* internal node */
          if (!st_lookup(fnode2pseudoOutput, (char *) Fnode_Regular(sib), (char **) &po)) {
              result = fprintf(fp,"%p ",Fnode_Regular(sib));
              if (result == EOF) return(0);
          }
          else {
              result = fprintf(fp,"%s ",po->name);
              if (result == EOF) return(0);
          }
      }
      else if (Fnode_Regular(sib)->value == BDS_FTREE_PSEUDO) { /* pseudo PI */
          result = fprintf(fp,"%s ",Fnode_Regular(sib)->name);
          if (result == EOF) return(0);
      }
      else { /* terminal node, print pi name */
          result = fprintf(fp,"%s ",PIvariables[Fnode_Regular(sib)->value]);
          if (result == EOF) return(0);
      }
  }
  lsFinish(fanin_list);

  /* Print functional node's name. In case of (a+b)' and (fg+f'h)', print newly
     allocated inverter pointer instead
  */
  if (Fnode_Operator(top) == BDS_BDD_OR && Fnode_Negative(top) ||
               Fnode_Operator(top) == BDS_BDD_MUX && Fnode_Negative(top)) {

      result = fprintf(fp,"%p\n",inverter);
      if (result == EOF) return(0);
  }
  else {
      if (dummy_node == NULL) { /* internal node */
          result = fprintf(fp,"%p\n",top);
          if (result == EOF) return(0);
      }
      else { /* PO node, print its name */
          result = fprintf(fp,"%s\n",dummy_node->name);
          if (result == EOF) return(0);
      }
  }

  return(1);

} /* end of bdsPrintOneNodePrintVarNames */

/**Function********************************************************************

  Synopsis    [Print out the function of this fnode]

  Description [Only AND and OR exist on the ftree. 
  Return 1 if successful, 0 otherwise]

  SideEffects [None]

  SeeAlso     [bdsDumpFtreeBlifOneNodePrintFunction()]

  Last date   [3/3/99]

*****************************************************************************/
static
int
bdsPrintOneNodePrintFunction(
  FILE *fp,
  FactorTreeNode *top)
{
  int result, i, j, num_fanins;
  FactorTreeNode *sib, *ctl, *inverter;
  FactorTreeNode *firstItem, *lastItem;
  BnetNode *po;
  lsList fanins;
  lsGen fanin_list;

  fanins = Fnode_Fanins(top);
  num_fanins = lsLength(fanins);
  fanin_list = lsStart(fanins);

  /* print functionality of this node */
  switch(Fnode_Operator(top)) {
  case BDS_BDD_AND:
      while (lsNext(fanin_list, (lsGeneric *)&sib, NULL) != LS_NOMORE) {

          result = fprintf(fp,"%c",\
               Fnode_Internal(sib)?(Fnode_IsComplement(sib)?'0':'1'):(Fnode_Negative(sib)?'0':'1'));
          if (result == EOF) return(0);
      }
      result = fprintf(fp," %c\n",Fnode_Negative(top)?'0':'1');
      if (result == EOF) return(0);
      break;
  case BDS_BDD_OR:
      /* The corresponding BLIF representation is a square, the negative case has
      ** been taken care of in bdsDumpFtreeBlifOneNodePrintVarNames().
      */
      i = 0;
      while (lsNext(fanin_list, (lsGeneric *)&sib, NULL) != LS_NOMORE) {
          for (j = 0; j < num_fanins; j++) {
              if (j == i) { /* print function */
                  result = fprintf(fp,"%c",\
                      Fnode_Internal(sib)?(Fnode_IsComplement(sib)?'0':'1'):(Fnode_Negative(sib)?'0':'1'));
                  if (result == EOF) return(0);
              }
              else { /* print - */
                  result = fprintf(fp,"-");
                  if (result == EOF) return(0);
              }
          }
          result = fprintf(fp," 1\n");
          if (result == EOF) return(0);

          i++;
      }
      break;
  default:
      printf("Unknown operator in bdsPrintOneNodePrintFunction()"); 
  }
  lsFinish(fanin_list);

  return(1);

} /* end of bdsPrintOneNodePrintFunction */

/**Function********************************************************************

  Synopsis    [Print out the function of this fnode]

  Description [Only AND and OR exist on the ftree. 
  Return 1 if successful, 0 otherwise]

  SideEffects [None]

  SeeAlso     [bdsDumpFtreeBlifOneNodePrintFunction()]

  Last date   [3/3/99]

*****************************************************************************/
static
int
bdsIfDuplicate(
  lsList newPIs, 
  FactorTreeNode *fnode,
  char **PIvariables)
{
  int retVal = 0;
  lsGen gen;
  FactorTreeNode *list_element;
  char *s1, *s2;

  gen = lsStart(newPIs);

  while (lsNext(gen, (lsGeneric *)&list_element, NULL) != LS_NOMORE) {
      if (Fnode_Regular(fnode)->value == BDS_FTREE_PSEUDO) { /* Pseudo PIs */
	  if (Fnode_Regular(list_element)->value == BDS_FTREE_PSEUDO) {
	      s1 = Fnode_Regular(list_element)->name;
	      s2 = Fnode_Regular(fnode)->name;
	      retVal = !strcmp(s1, s2);
	      if (retVal != 0) goto endgame;
          }
	  else {
	      retVal = 0;
	  }
      }
      else { /* Normal PIs */
	  if (Fnode_Regular(list_element)->value != BDS_FTREE_PSEUDO) {
	      s1 = PIvariables[Fnode_Regular(list_element)->value];
	      s2 = PIvariables[Fnode_Regular(fnode)->value];
	      retVal = !strcmp(s1, s2);
	      if (retVal != 0) goto endgame;
          }
	  else {
	      retVal = 0;
	  }
      }
  }

endgame:

  lsFinish(gen);
  return(retVal);

} /* end of bdsIfDuplicate */
