/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CHeaderFile*****************************************************************

  FileName    [lopt.h]

  PackageName [BDS]

  Synopsis    [BDD decomposition-based logic optimization program]

  Description []

  SeeAlso     []

  Author      [Anda Congguang Yang]

******************************************************************************/

#ifndef _BDS
#define _BDS

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#ifdef _MSC_VER
#define _WINDOWS
#endif

#include "build.h"
#include "bnet.h"
#include "array.h"
#include "list.h"
#include "st.h"
#include "cuddInt.h"
#include <time.h>

#ifdef _WINDOWS
#include "winfake.h"
#else
#include <sys/utsname.h>
#endif

/*---------------------------------------------------------------------------*/
/* subminor history 
 *  01: at the BDS_VERBODE_MOST level, BDD size before and after reorder reported
 *      (manual loop discovery)
 *  02: -xdelta and x-decomp disqualification introduced
 *  03: -xhardcore backported from bds-pga, -dir introduced
 *  04: -xhardcore option removed, -sid -ged -bhc -xhc -cft introduced
 *  05: bug in BDS_GetFileHead corrected
 *  06: bnet.c, loptUtil.c: Cudd_DumpBlif usage conforming CUDD 2.4.2
 *      eliminate.c: BDS_EliminateNodeOrNot crashed on constant outputs
 *  07: bnet.c: CR LF resistance in BLIF import
 *      lopt.c: BDS_CheckTerminal(), bug overriding ftnode value
 *      loptUtil.c: BDS_GetFileHead() changed to BDS_GetEntireFileHead(), BDS_GetFileHead() changed
 *      export.c: error message on opening output file enhanced 
 *      local.c: bdsConvertDummy2RealRecursive() no conversion on constant nodes
 *      ftree.c: BDS_ExtractSharingOnFtreeRecursive() proper treatment of constant nodes
 *  08: local.c: trivial mistake in the previous mod
 *  09: export.c: bdsDumpFtreeBlifOneNodelsList(): constant ftree nodes detected
 *      export.c: bdsDumpFtreeBlifOneNodelsList(): inputs to PO nodes from Bnet not PIvariables 
 *      export.c: bdsDumpFtreeBlifOneNode(): constant ftree nodes detected; 
 *              constant node semantics taken over from Bnet_PrintNetwork()
 *      export.c: bdsDumpFtreeBlifOneNode(): outputs connected to constant generators 
 *              output these generators and mark them visited. bdsDumpFtreeBlif() has
 *              to erase the marks; net->nodes useless (chain corrupted - where?)
 *      export.c: bdsDumpFtreeBlifOneNode(): inputs to PO nodes from Bnet not PIvariables 
 *      lopt.c: BDS_CheckTerminal(), bug in constant bdd detection (introduced in 1.2.07)
 *      prepare.c: bdsBuildVarNameAssocLocal() only PIs and POs(why?) collected into PIvariables
 *      bnet.[ch]: bnodeTypes added for debugging purposes
 *	bnet.h: constant node semantics recorded in comments
 *	main.c: bdsDumpFtreeBlif() in export.c assumes that net->outputs[i] corresponds to FTNode[i];
 *		enforce this in FTNode construction
 *      loptUtil.c: BDS_CreateRunDir() puts the entire year into the name, not years since 1900;
 *              considered too cryptic after 2000
 *  10: prepare.c: bdsBuildVarNameAssocLocal(): BDD variables collected into PIvariables
 * 		only when the Bnode is a PI, PO, or internal (not constant). Bug introduced 
 *		in 1.2.09.
 *	main.c: -v option introduced
 *  11: main.c: alter poName2poIndex to store pointers into FTreeNode not indices (easier 64 bit portability)
 *      export.c, debug.c, cutptl.c: alter fprintf() formats to use '%p' not '0x%x'. 
 *		Do not cast its operand to unsigned. Affects output produced by BDS_DumpBddBlif() -
 *		some names have '0x' in front of them. should not affect BLIF conformance.
 *  12: export.c: bdsDupFtreeBlifOneNode(),  bdsDupFtreeBlifOneNodelsList():
 *              code for constant primary outputs and buffers/inverters between 
 *              primary inputs and outputs rewritten again, with the assumption that internal
 *              constant generators have been deleted by constant propagation
 *              (closer to the original, deleting most of the code from 1.2.09)
 *      export.c: bdsDumpFtreeBlif(), BDS_DumpFtreeBlif(): any node is collected as output
 *              if it has an entry in FactorTreeNode2BnetNode (affects intermediate BLIF outputs)
 *      bulid.c: BDS_BuildLocalNodeBDD(), BuildGlobalNodeBDD(): employ the DFS manner
 *              of BDD construction to do constant propagation
 *      main.c: in the main processing loop, skip BDDs that correspond to internal constant
 *              generators - they are not needed any more
 *      debug.h: new file, debug function prototypes (also from lopt.h), debug control
 *              (compile-time)
 *      debug.c: functions for tabelating factorization tree nodes
 *      lopt.h: debug-related macros and prototypes moved to debug.h
 *      main.c, ftree.c, export.c, lopt.c, sweep.c, minimize.c: altered #ifdef DEBUG to #if DEBUG_ACTIVE(ddd)
 *              (see debug.h)
 *  13: bnet.c: unreasonable limitation on list size in readList(). Code rewritten,
 *		based on REALLOC.
 *  14: winfake.h, winfake.c: replacement functions for windows compilation
 *      export.c: BDS_DumpSingleFTree(): added code for constant generator
 *              as an emergency measure (the EQN files can be invalid)
 *  15: maintenance
 *  16: lopt.c, loptUtil.c, cutptl.c: unreasonable limitation on string size. 
        Code rewritten. 
        build.h, lopt.c, main.c: timeout in BDS_BddDecomposeMain
        manpage: -delta usage clarified, -time introduced
 */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define BDS_VERSION "BDS Version #1.2.16, Release date Aug 2024"

/* Constants for operator field in factor tree */
#define BDS_BDD_INV 0
#define BDS_BDD_AND 1 	
#define BDS_BDD_OR 2
#define BDS_BDD_XOR 3
#define BDS_BDD_XNOR 4
#define BDS_BDD_MUX 5 

#define BDS_NUMBER_OPERATOR 6 

/* Decomposition type */
#define BDS_BDD_DECOMP_NONE 0	
#define BDS_BDD_DECOMP_CONJ 1
#define BDS_BDD_DECOMP_DISJ 2
#define BDS_BDD_DECOMP_BOTH 3
#define BDS_BDD_DECOMP_X    4
#define BDS_BDD_DECOMP_B    5

/* Return values of BDD decomposition */
#define BDS_BDD_NO_DECOMP     0xFFF0
#define BDS_BDD_NO_MINIMI     0xFFF1
#define BDS_DECOMP_DISQUALIFY 0xFFF2

/* Cmd for BDS_Stats */
#define BDS_INIT	  	0xFFF0
#define BDS_BDD_DECOMP   	0xFFF2
#define BDS_XHARDCORE	  	0xFFF4
#define BDS_BHARDCORE 	        0xFFF5
#define BDS_SHARED_FNODE 	0xFFF6
#define BDS_CLOSE	  	0xFFF8
#define BDS_VERIFY	  	0xFF0F
#define BDS_ONE_DOMI	  	0xFF00
#define BDS_ZERO_DOMI	  	0xFF02
#define BDS_X_DOMI	  	0xFF04
#define BDS_G_DOMI	  	0xFF06
#define BDS_LIT_COUNT		0xFF08
#define BDS_COFACTOR		0xFF0A
#define BDS_FTREE_OPERATOR      0xFF0C
#define BDS_REORDER		0xFF0D	/* JS */
#define BDS_HACKOUT		0xFF0E	/* JS */

/* Verbosity level */
#define BDS_VERBOSE_LESS 0
#define BDS_VERBOSE_MORE 1 
#define BDS_VERBOSE_MOST 2

/* Default startup parameters */
#define BDS_BDD_STARTUP_CACHE  10*1024
#define BDS_BDD_STARTUP_MEMORY 10*1024*1024   
#define BDS_BDD_REORDER_LIMIT  200    /* The largest BDD on which exact reordering can apply. */

/* BDD minimization method */
#define BDS_BDD_RESTRICT	0
#define BDS_BDD_LICOMPACTION	1

/* ftree node */
#define BDS_FTREE_INTERNAL	0xFFFF /* Internal node of a ftree */
#define BDS_FTREE_ONE		0xFFF1 /* Constant 1 on a ftree */
#define BDS_FTREE_ZERO		0xFFF0 /* Constant 0 on a ftree */
#define BDS_FTREE_PSEUDO        0xFFF2 /* Used for cut PTLs */

/* Weight-realted constants */
#define BDS_HUGE_NUMBER 0xFFFF
#define ALPHA 0.5		       /* Weight constant for BDD decomp cost function(GD) */
#define BETA 0.5		       /* Weight constant for BDD decomp cost function(XD) */
#define BDS_DISQUALIFY_THRESHOLD 0.80 /* Threshold of cost to disqualify a decomposition */
#define BDS_DISQUALIFY_GXD_RATIO 0.95 /* The threshold to disqualify decomposition based on gx-d */

/* BDD balancing related parameters */
#define BDD_BALANCE_FACTOR 0.5
#define BDD_BALANCE_THRESHOLD 0.3

/* X decomp loop protection JS */
#define BDS_NO_DELTA 0x7FFFFFFF

/* decomp methods JS */
#define BDS_METH_SID 0      /* simple dominators - algebraic */
#define BDS_METH_GED 1      /* generalized dominators */
#define BDS_METH_BHC 2      /* branch decomp - b hardrcore */
#define BDS_METH_XHC 3      /* x decomp - x hardcore */
#define BDS_METH_CFT 4      /* cofactor on top */

/* decomp control in options->methods */
#define BDS_CTRL_SID (1 << BDS_METH_SID)
#define BDS_CTRL_GED (1 << BDS_METH_GED)
#define BDS_CTRL_BHC (1 << BDS_METH_BHC)
#define BDS_CTRL_XHC (1 << BDS_METH_XHC)
#define BDS_CTRL_CFT (1 << BDS_METH_CFT)

#define BDS_CTRL_ALL (BDS_CTRL_SID | BDS_CTRL_GED | BDS_CTRL_BHC | BDS_CTRL_XHC | BDS_CTRL_CFT)

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/
typedef int boolean;

typedef struct FactorTreeNode {
    int 	operator;	/* operator on two children,  +, *, ^ and @ */
    int		value;		/* Value of this FTNode, i.e., the index of the variable */
    char	*name;		/* Used for cut PTLs */
    DdNode	*dd;		/* Corresponding global BDD */
    DdNode	*local;	        /* Local BDD, supports are intermediate variables */
    int		var;		/* Local variable index */
    int		count;		/* Reference count on this node */
    int 	visited;	/* Used for factor tree traversal */
    int		polarity;	/* Complementary (0) or not (1) */
    struct FactorTreeNode *siblings[3];	/* Operands on this fnode, used to hold intermediate results */
    struct FactorTreeNode *control;     /* Used to store control signal after collapse fnodes */
    lsList	fanins;		/* fanin nodes */
    lsList	fanouts;	/* fanout nodes */
} FactorTreeNode;

typedef struct bddNode {
    int            index;       /* Index of the node */
    DdNode         *dd;         /* BDD corresponds to this node */
    struct bddNode *t_nd;       /* THEN child */
    struct bddNode *e_nd;       /* ELSE child */
} bddNode;

typedef struct bddPool {
    int         *order;         /* Variable order when this BDD is dumped */
    bddNode     *bnode;
    int		numVar;		/* Number of vars on the BDD */
} bddPool;

typedef struct SopPool {	/* Struct used to store tmp result of a BDD decomposition */
    int 	size;		/* # of variables in a bdd mgr */
    int 	*order;		/* variable order, used to rebuild the BDD from SOP */
    array_t 	*sop;		/* miniterms stored here */
} SopPool;

typedef struct SQueue {		/* A naive queue struct. */
  long int 	head;
  long int 	tail;
  array_t 	*queue;
} SQueue;

typedef struct decompCost {	/* Cost associated with a decomposition */
  float 	cost;		/* Cost for this decomposition */
  int		numQR;		/* # of nodes on the decomposed BDD */
  int		numD;		/* # of nodes on the dominator */
  int 		operator;	/* operator for this decomposition */
  int 		level;		/* The level on which the best decomp is found */
  int 		decompType;
  bddPool	*decomposed;
  bddPool	*dominator;
} decompCost;

typedef struct XdecompCost {	/* Cost associated with a X decomposition */
  float 	cost;		/* Cost for this decomposition */
  int 		operator;	/* operator for this decomposition */
  int 		decompType;
  DdNode	*f;
  DdNode 	*g;
  DdNode 	*h;
} XdecompCost;

typedef struct edgeMark {	/* Hash table used to store edge property */
  int 		decompType;	/* Type of possible decomp on this level. */
  int 		numSigmaOne;	/* Number of Sigma_1 on this level */
  int 		numSigmaZero;	/* Number of Sigma_0 on this level */
  int		numNegative;	/* Number of negative edges on and above this level */
  DdNode 	*bottom;	/* The lower end of an edge */
  st_table 	*table;		/* Store the lower end (lBdd) of edges */
} edgeMark;
 
typedef struct bddStats {       /* BDD decomposition statistical monitor */
   int           inVar;         /* # of variables on the in coming BDD */
   int           inSize;        /* Size of incoming BDD */
   int           dpVar;
   int           dpSize;
   int           dmVar;
   int           dmSize;
   int		 auxVar;	/* Only used for branch decomposition */
   int		 auxSize;
   int           level;
   int           decompType;
   float         cost;
} bddStats;

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/* ftree.c */
#define Fnode_Regular(node) ((FactorTreeNode *)((unsigned long)(node) & ~01))
#define Fnode_Complement(node) ((FactorTreeNode *)((unsigned long)(node) | 01))
#define Fnode_Flip(node) ((FactorTreeNode *)((unsigned long)(node) ^ 01))
#define Fnode_IsComplement(node) ((int) ((long) (node) & 01))
#define Fnode_Ref(node) (Fnode_Regular(node)->count ++)
#define Fnode_Deref(node) (Fnode_Regular(node)->count --)
#define Fnode_Internal(node) (Fnode_Regular(node)->value == BDS_FTREE_INTERNAL)
#define Fnode_Negative(node) (Fnode_Regular(node)->polarity == 0)
#define Fnode_Inverse(node) (Fnode_Regular(node)->polarity = 1 - Fnode_Regular(node)->polarity)
#define Fnode_Operator(node) (Fnode_Regular(node)->operator)
#define Fnode_Fanins(node) (Fnode_Regular(node)->fanins)
#define Fnode_Fanouts(node) (Fnode_Regular(node)->fanouts)

/* store.c */
#define DumpNode_Regular(node) ((bddNode *)((unsigned long)(node) & ~01))
#define DumpNode_Complement(node) ((bddNode *)((unsigned long)(node) | 01))
#define DumpNode_Flip(node) ((bddNode *)((unsigned long)(node) ^ 01))
#define DumpNode_IsComplement(node) ((int) ((long) (node) & 01))

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

/* main.c */
EXTERN DdManager * startCudd ARGS((bdsOptions *option, int nvars));

/* lopt.c */
EXTERN int BDS_DDDecompose ARGS((DdManager*,DdNode*,bdsOptions*,FactorTreeNode**));

/* queue.c */
EXTERN void BDS_Q_Init ARGS((SQueue*));
EXTERN void BDS_EQ ARGS((SQueue*, bddPool*));
EXTERN bddPool * BDS_DQ ARGS((SQueue*));
EXTERN void BDS_Q_Quit ARGS((SQueue*));

/* minimize.c */
EXTERN DdNode * BDS_BddMinimization ARGS((DdManager*,DdNode*,DdNode*,int,bdsOptions*));

/* ftree.c */
EXTERN int BDS_FTreeProcessing ARGS((DdManager*,BnetNetwork*,FactorTreeNode**,bdsOptions*,char**,st_table*));
EXTERN void BDSFreeFTree ARGS((DdManager*,FactorTreeNode**));
EXTERN int bdsCollectFNodes ARGS((st_table*,FactorTreeNode*));
EXTERN int bdsCollectFNodeslsList ARGS((st_table*,FactorTreeNode*));

/* export.c */
EXTERN int BDS_SynthesisPresentation ARGS((BnetNetwork*,FactorTreeNode**,char**,st_table*,st_table*,bdsOptions*,int));

/* cutptl.c */
EXTERN int BDS_CutPtlFromGeneralNetlist ARGS((BnetNetwork*,FactorTreeNode**,char**,st_table*,bdsOptions*));

/* phase.c */
EXTERN int BDS_FtreePhaseAssignment ARGS((FactorTreeNode**,bdsOptions*));

/* store.c */
EXTERN bddPool * BDS_DumpBddPool ARGS((DdManager*,DdNode*));
EXTERN DdNode * BDS_BuildDDFromBddPool ARGS((DdManager*,bddPool*,int));
EXTERN bddPool * BDS_DumpBddPoolAtLevel ARGS((DdManager*,DdNode*,st_table*,edgeMark*,int));
EXTERN void BDS_FreeBddPool ARGS((DdManager*,bddPool*));
EXTERN DdNode * bdsBuildResubBddFromBddPool ARGS((DdManager*,bddPool*));
EXTERN bddPool * bdsDumpResubBddPool ARGS((DdManager*,DdNode*,DdNode*,DdNode*));

/* debug.c */
#include "debug.h"

/* loptUtil.c */
EXTERN edgeMark ** BDS_MarkEdges(DdManager*,DdNode*,int*,int);
EXTERN DdNode* BDS_BuildDDFromSOP ARGS((DdManager*,SopPool*));
EXTERN SopPool * BDS_DumpSOP ARGS((DdManager*,DdNode*));
EXTERN void BDS_FreeEdgeProperty ARGS((int*,edgeMark**,int));
EXTERN void BDS_FreeSOP ARGS((SopPool*));
EXTERN array_t * BDS_Support ARGS((DdManager*,DdNode*));
EXTERN SopPool * BDS_DumpSopAtLevel ARGS((DdManager*,DdNode*,edgeMark*,int));
EXTERN int* BDS_SupportArray ARGS((DdManager*,DdNode*,int*));
EXTERN char * BDS_ReadBdd ARGS((DdManager*,DdNode*,char**));
EXTERN char * BDS_DumpBddBlif ARGS((DdManager*,DdNode*,char**));
EXTERN int BDS_Reorder ARGS((DdManager*,BnetNetwork*,bdsOptions*));
EXTERN void BDS_DumpDD ARGS((DdManager*,BnetNode*,BnetNetwork*,bdsOptions*,int));
EXTERN void BDS_PrintStatus ARGS((DdManager*,BnetNode*,int));
EXTERN void BDS_CheckRefCount ARGS((DdManager*,BnetNode*));
EXTERN char * BDS_CreateFname ARGS((bdsOptions*,char*));
EXTERN char * BDS_GetEntireFileHead ARGS((char*));
EXTERN char * BDS_GetFileHead ARGS((char*));
EXTERN void BDS_Stats ARGS((int,bddStats*,long,char*,char**,DdManager*,bdsOptions*));
EXTERN void BDS_WriteFnode ARGS((FactorTreeNode*,int,int,int));
EXTERN void BDS_UpdateBddMonitor ARGS((DdManager*,bddStats*,DdNode*,DdNode*,DdNode*,DdNode*,int,float,int));
EXTERN int BDS_GetNegative ARGS((DdManager*,DdNode*,int));
EXTERN st_table * BDS_CountRef ARGS((DdManager*,DdNode*));
EXTERN int BDS_TestBalance ARGS((DdManager*,DdNode*));

#endif /* _BDS */
