/*************************************************************************
	-----------------------------------------------------
           		Copyright 2002, 
	   University of Massachusetts, Amherst. 
	-----------------------------------------------------

This package was created at the University of Massachusetts, Amherst. 
The University of Massachusetts makes no warranty about the suitability 
of this software for any purpose. It is presented on an "as is" basis.

CONDITIONS:

The use of software for commercial purposes is strictly forbidden
without a written consent from its authors or the University of 
Massachusetts, Amherst.

**************************************************************************/
/**CFile***********************************************************************

   FileName    [main.c]

   PackageName [BDS]

   Synopsis    [Main program for the BDS program.]

   Description []

   SeeAlso     []

   Author      [Anda Congguang Yang]

******************************************************************************/

#include "lopt.h"
#include "enhance.h"
#include "sweep.h"
#include "prepare.h"
#include "sharing.h"
#include "eliminate.h"
#include "local.h"
#include "verify.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static bdsOptions * mainInit ARGS(());
static void bdsReadOptions ARGS((int argc, char **argv, bdsOptions *option));
static void freeOption ARGS((bdsOptions *option));

/* move to lopt.h
EXTERN DdManager * startCudd ARGS((bdsOptions *option, int nvars));
*/

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Main program for BDS]

  Description []

  SideEffects [None]

  SeeAlso     []

  LastDate    [5/16/99]

******************************************************************************/
int
main(
  int  argc,
  char ** argv)
{
    bdsOptions	*option;        /* Command line holder */
    FILE	*fp;		/* network file pointer */
    BnetNetwork	*net = NULL;	/* network */
    int	result;		/* stores the return value of functions */
    BnetNode	*node;		/* auxiliary pointer to network node */
    int	i;		/* loop index */
    int	j;		/* loop index */
    FactorTreeNode **FTreeNode;	/* an array of PO nodes in the factor tree */
    FactorTreeNode **fptr;	/* ptr into the FTNode vector */
    FactorTreeNode **fnext;	/* next free item in the FTNode vector */
    DdManager   *bddmgr;	/* BDD manager */
    DdManager   *decompBddmgr;	/* Dummy BDD manager */
    char	**PIvariables;	/* association list of PI name and index in a BDD */
    char	*recordDump;    /* File to which reordering information is dumped */
    FILE	*recdfp;	/* file handle for recordDump */
    int	globalReady=0;  /* global BDDs ready flag */
    int	localOrGlobal=0;/* Use local(0) or global(1) BDD to do lopt */
    bddPool	**bddPoolArray, **q; /* All to be decomposed BDD array */
    st_table 	*bddPoolBnodeAssoc;       /* bddPool and BnetNode assoc */
    st_table	*factorTreeNode2BnetNode; /* FTNode and BnetNode assoc */
    st_table    *poName2poIndex;	  /* output name -> ouput index assoc (inverted net->outputs); */
    DdNode	*f;		/* to be decomposed bdd holder */
    int		*dummy2realId;  /* substitute Id and real Id assoc */
    long BDDDecompStartTime, BDDDecompEndTime, BDDDecompTotalTime;
    long netProcesStartTime, netProcesEndTime, netProcesTotalTime;

    /* Command line parser */
    option = mainInit();
    (void) bdsReadOptions(argc,argv,option);

    /* Read the Boolean network */
    fp = fopen(option->file, "r");
    if (fp == NULL) {
        fprintf(stderr, "Can not open file: %s\n", option->file);
	exit(2);
    }
    net = Bnet_ReadNetwork(fp,option->verb);
    (void) fclose(fp);
    if (net == NULL) {
	(void) fprintf(stderr,"Syntax error in %s.\n",option->file);
	exit(2);
    }
    if (net->nlatches != 0) {
	printf("Sorry, currently BDS only handles combinational circuits\n\n");
	exit(0);
    }
    if (option->verb > BDS_VERBOSE_MOST) {
	Bnet_PrintNetwork(net);
    }

    /* Enhance the network. Internal Boolean network data structure in CUDD is
    ** augumented to enable network partitioning operation. 
    */
    result = BDS_BnetEnhance(net,option);
    if (result == 0) exit(2);

    /* Initialize bdd manager. */
    bddmgr = startCudd(option,net->npis);
    if (bddmgr == NULL) { exit(2); }

#ifndef DEBUG
    /* Redirect the message related to variable reordering to a record file 
    */
    recordDump = BDS_CreateFname(option,"reorder.stat");
    recdfp = fopen(recordDump, "w");
    if (recdfp == NULL) {
        printf("Can not create files !\n");
	exit(2);
    }
    Cudd_SetStdout(bddmgr, recdfp);
#endif

    netProcesStartTime = util_cpu_time();

    /* If -useloc is specified, don't build global BDD, this will save the time
    ** to build and then abort global construction on large test cases.
    */
    if (!option->useloc == TRUE) {
        /* Build global BDDs. Turn on dynamic variable reordering in case the global
        ** BDDs are huge. If globalReady == 0, build global BDDs fails. 
        */
        Cudd_AutodynEnable(bddmgr,CUDD_REORDER_SIFT);
        result = BDS_BuildGlobalBDD(bddmgr,net,option,&globalReady);
        if (result == 0) exit(2);

        /* Eliminate node based on global BDDs */
        if (globalReady == 1 && option->glbEliminate) {
            result = BDS_EliminateBasedOnGlobal(bddmgr,net,option);
	    if (result == 0) exit(2);
        }
    }

    /* If -useglb is specified and the global BDD are unable to be built, quit.
    */
    if (option->useglb && globalReady == 0) {
        printf("\nWarning: The size of global BDDs is > %d which is current threshold.\n", 
								option->globalLimit);
	printf("\t Use \"-useloc\" is highly recommended.\n\n");
	printf("\t If you really want to see the synthesis results based on global BDDs,\n");
	printf("\t please use \"-globalLimit number\" to specify a larger threshold.\n\n\n");
        Cudd_Quit(bddmgr);
	exit(0);
    }

    /* Automatic dynamic variable reordering is disabled thru all. Variable reordering
    ** when local BDDs are built may cause wrong BDD nodes recorded on each Bnode. 
    */
    Cudd_AutodynDisable(bddmgr);

    /* If -useglb is specific, don't build local BDDs, all BDD will collapse
    ** into the PO nodes and the final circuit is synthesized from global BDDs.
    */
    if (!option->useglb == TRUE) {
        /* Build local BDDs */
        result = BDS_BuildLocalBDD(bddmgr,net,option);
        if (result == 0) exit(2);

        /* Sweep the Bnetwork to remove constant and duplicate bnodes */
	result = BDS_SweepBnetwork(bddmgr,net,option);
	if (result == 0) exit(2);

        /* Eliminate nodes based on elimination gain threshold. If rodelim is specified
        ** the network is first eliminated using non-reorder, then use reorder-based eliminate. 
        */
        result = BDS_EliminateBasedOnLocal(bddmgr,net,option,globalReady,0,recdfp);
        if (result == 0) exit(2);
        if (option->rodelim) {
            result = BDS_EliminateBasedOnLocal(bddmgr,net,option,globalReady,1,recdfp);
            if (result == 0) exit(2);
        }

        /* Sweep the Bnetwork to remove constant and duplicate bnodes */
	result = BDS_SweepBnetwork(bddmgr,net,option);
	if (result == 0) exit(2);

        if (option->dumpnet == TRUE) netPrint(net,option);

        /* Network partition is verified only if global BDDs can be built. */
        if (option->verify == TRUE && globalReady) {
	    result = bdsVerifyPartition(bddmgr,net,option);
	    if (result == 0) {
	        printf("Fatal error in partition");
	        exit(2);
	    }
        }
    }

    /* Decide whether global or local BDDs are used for synthesis */
    if (option->useglb) { /* Use global BDDs */
        localOrGlobal = 1;
    }
    else if (option->useloc) { /* Use local BDDs */
        localOrGlobal = 0;
    }
    else if (globalReady) { /* Compare global and local, then decide */
        localOrGlobal = bdsGlobalOrLocalBDD(bddmgr,net,option,recdfp);
    }

    /* Sharing extraction based on multi-BDDs, new bnodes are created to hold 
    ** those shared logic. All BDDs on each node are stored in bddPool format 
    ** and wait to be decomposed.
    */
    bddPoolBnodeAssoc = st_init_table(st_ptrcmp,st_ptrhash);
    if (localOrGlobal) { /* Use global BDD */
        printf("Global BDDs are used for synthesis\n\n");

	/* Sharing extraction enginer entry */
	result= BDS_SharingExtractionGlobal(bddmgr,net,option);
	if (result == 0) {
	    printf("Sharing extraction (global) fails !\n\n");
	    exit(2);
	}
	/* Store all global BDDs on the PO nodes */
	bddPoolArray = bdsStoreBDDInBddPoolGlobal(bddmgr,net,option,bddPoolBnodeAssoc);
	if (bddPoolArray == NULL) exit(2);

	/* Free all BDDs */
	result = bdsFreeAllBDDs(bddmgr,net,option,globalReady);
        if (result == 0) exit(2);

        /* Build varId and char* name association */
	PIvariables = bdsBuildVarNameAssocGlobal(bddmgr,net,option);
	if (PIvariables == NULL) exit(2);
    }
    else { /* Use local BDD */
        printf("Local BDDs are used for synthesis\n\n");

	/* Sharing extraction enginer entry */
	result= BDS_SharingExtractionLocal(bddmgr,net,option,recdfp);
	if (result == 0) {
	    printf("Sharing extraction (local) fails !\n\n");
	    exit(2);
	}
	/* Store all local BDDs */
	bddPoolArray = bdsStoreBDDInBddPoolLocal(bddmgr,net,option,bddPoolBnodeAssoc);
	if (bddPoolArray == NULL) exit(2);

	/* Free all BDDs */
	result = bdsFreeAllBDDs(bddmgr,net,option,globalReady);
        if (result == 0) exit(2);

        /* Build varId and char* name association */
	PIvariables = bdsBuildVarNameAssocLocal(bddmgr,net,option);
	if (PIvariables == NULL) exit(2);
    }

    /* Allocate a root fnode for each BDD */
    i = 0;
    for (q = bddPoolArray; *q; q++) {
        i++;
    }
    FTreeNode = ALLOC(FactorTreeNode *, i + 1);
    FTreeNode[i] = NULL;

    netProcesEndTime = util_cpu_time();
    netProcesTotalTime = netProcesEndTime - netProcesStartTime;
    BDDDecompStartTime = util_cpu_time();

    /* Start decomposing BDDs, one by one */
    factorTreeNode2BnetNode = st_init_table(st_ptrcmp,st_ptrhash);
    printf("Decomposing BDDs ");
#if DEBUG_ACTIVE(DEBUG_COLLECT_FTREE)
    printf("\n");
#endif
    (void) fflush(stdout);

#ifndef DEBUG
    fprintf(recdfp, "\n\n<<<<<<<<<<  Decomposing BDDs  >>>>>>>>>>\n\n");
#endif
    /* bdsDumpFtreeBlif assumes that net->outputs[i] corresponds
     * to FTreeNode[i], which holds only when global BDDs are used.
     * In local BDDs mode (localOrGlobal == 0), force this correspondence.
     */

    if (localOrGlobal == 0) {
        /* invert net->outputs */
        poName2poIndex = st_init_table (strcmp, st_strhash);
        if (poName2poIndex == NULL) {
            printf ("\nPO table init failed\n");
            exit(2);
        }
	for (i=0; i<net->npos; i++) {
	    if (st_add_direct(poName2poIndex, net->outputs[i], (char*)(FTreeNode + i))
				== ST_OUT_OF_MEM) {
	        printf("\nOut of memory !\n");
	        exit(2);
	    }
	}
	fnext = FTreeNode + net->npos;
    } else {
        fnext = FTreeNode;
    }
    
    
    for (q = bddPoolArray; *q; q++) {
	/* get BnetNode for a given BDD pool record */
	if (!st_lookup(bddPoolBnodeAssoc, (char *) *q, (char **) &node)) {
	    fail("Fatal error in bddPoolBnodeAssoc!");
	}

	/* Record FactorTreeNode vs. BnetNode correspondency */
	/* get FTreeNode pointer */
	/* the algorithm should work for global as well as local BDDs,
	 * but, in the tangle of undocumented dependencies, this is the safer way. JS */
	if (localOrGlobal == 0) {
	    if (node->type == BNET_CONSTANT_NODE) {
	        if (st_lookup (poName2poIndex, node->name, &fptr) == 0) {
	             continue;  /* this is an internal constant generator, not needed */
	        }
	    } else if (node->type == BNET_OUTPUT_NODE) {
	        if (st_lookup (poName2poIndex, node->name, &fptr) == 0) {
	             fptr = fnext++;    /* probably cannot happen */
	        }
	    } else {
	        fptr = fnext++;
	    }
	} else {
	    fptr = fnext++;
	}
	
	/* construct a FactorTreeNode */
	(*fptr) = ALLOC(FactorTreeNode, 1); 
	DBG_MEMSET ((*fptr), 0, sizeof (FactorTreeNode));
	if (st_add_direct(factorTreeNode2BnetNode, (char *) (*fptr), (char *) node) 
				== ST_OUT_OF_MEM) {
	    printf("\nOut of memory !\n");
	    exit(2);
	}
	
        *fnext = NULL;

#if DEBUG_ACTIVE(DEBUG_COLLECT_FTREE)
	printf ("collecting %s at %p as FTnode %p\n", node->name, fptr, *fptr);
#endif
        decompBddmgr = startCudd(option,0);
        if (decompBddmgr == NULL) { exit(2); }

#ifndef DEBUG
        Cudd_SetStdout(decompBddmgr, recdfp);
#endif

        /* Build dummy BDD on this node */
	f = BDS_BuildDDFromBddPoolLocal(decompBddmgr,*q,&dummy2realId,bddmgr->size,option);
	if (f == NULL) exit(2);

	/* Main BDD decomposition engine entry */
	result = BDS_DDDecompose(decompBddmgr, f, option, fptr);
        Cudd_Quit(decompBddmgr);

	if (result == 0) {
	    printf("fail !\n");
	    exit(2);
	}

	/* Convert dummy Ids back to real Ids */
	result = bdsConvertDummy2Real(*fptr, dummy2realId, option);
	if (result == 0) exit(2);

    }
    printf("Done\n\n");

#if DEBUG_ACTIVE(DEBUG_TABELATE_FTREE)
    dumpFTreeArrayBnodePI (stdout, FTreeNode, factorTreeNode2BnetNode, PIvariables);
#endif
    /* Factoring trees processing engine entry */
    result = BDS_FTreeProcessing(bddmgr,net,FTreeNode,option,PIvariables,factorTreeNode2BnetNode);
    if (result == 0) {
	printf("Final ftree processing fails !\n");
	exit(2);
    }

    /* Verfiy the final synthesis results */
    if (option->verify == TRUE) {
        result = bdsVerifySynthesis(bddmgr,net,FTreeNode,factorTreeNode2BnetNode,\
	              bddPoolArray,bddPoolBnodeAssoc,option);
        BDS_Stats(BDS_VERIFY,NULL,result,"after",NULL,NULL,NULL);
    }

    st_free_table(bddPoolBnodeAssoc);
    st_free_table(factorTreeNode2BnetNode);

    /* Free factoring tree */
    (void) BDS_FreeFTree(bddmgr, FTreeNode);
    FREE(FTreeNode);

#ifndef DEBUG
    fclose(recdfp);
#endif

    FREE(FTreeNode); FREE(PIvariables); 

    for (q = bddPoolArray; *q; q++) {
	(void) bdsFreeBddPoolLocal(bddmgr, *q);
    }
    FREE(bddPoolArray);

    BDDDecompEndTime = util_cpu_time();
    BDDDecompTotalTime = BDDDecompEndTime - BDDDecompStartTime; 

    BDS_Stats(BDS_CLOSE,NULL,BDDDecompTotalTime,NULL,NULL,bddmgr,option);
    Cudd_Quit(bddmgr);

    /* Free Boolean network */
    (void) BDS_FreeNetworkAfterEliminate(net);

    printf("\nTotal time used to do network preprocessing = %s\n", util_print_time(netProcesTotalTime));
    printf("Total time used to do BDD decomposition = %s\n\n", util_print_time(BDDDecompTotalTime));

    freeOption(option);
    util_print_cpu_stats(stdout);

    exit(0);

} /* end of main */

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Set the default options]

  Description []

  SideEffects [none]

  SeeAlso     [bdsReadOptions]

******************************************************************************/
static bdsOptions *
mainInit(
   )
{
    bdsOptions	*option;

    /* Initialize option structure. */
    option = ALLOC(bdsOptions,1);
    option->initialTime    = util_cpu_time();
    option->file           = NULL;
    option->cacheSize      = BDS_BDD_STARTUP_CACHE;
    option->maxMemory      = BDS_BDD_STARTUP_MEMORY;
    option->slots          = CUDD_UNIQUE_SLOTS;
    option->order          = PI_PS_FROM_FILE;
    option->orderfile      = NULL;
    option->reordering     = CUDD_REORDER_NONE;
    option->maxGrowth      = 20;
    option->verb           = BDS_VERBOSE_MORE;

    /* Options related to BDS */
    option->useloc	       = FALSE;
    option->useglb	       = FALSE;
    option->ethred	       = BDS_ELIMINATE_THRESHOLD;
    option->autodyn	       = FALSE;
    option->rodelim	       = FALSE;
    option->globalLimit	   = BDS_GLOBAL_BDD_LIMIT;
    option->glbEliminate   = FALSE;
    option->verify         = FALSE;
    option->effort	       = 1;
    option->limitSize 	   = BDS_BDD_REORDER_LIMIT;
    option->miniMethod     = BDS_BDD_RESTRICT;
    option->dumpbdd	       = FALSE;
    option->dumpblif	   = TRUE;
    option->dumpftree      = TRUE;
    option->dumpfdot       = FALSE;
    option->dumpnet	       = FALSE;
    option->cutptl         = FALSE;
    
    option->timeout        = BDS_NO_TIMEOUT;    /* default is no timeout */
    option->methods        = BDS_CTRL_ALL;      /* default is to use all methods */
    option->xdelta         = BDS_NO_DELTA;      /* default is no limit on BDD enlargement */
    option->dir            = NULL;

    return(option);

} /* end of mainInit */

/**Function********************************************************************

  Synopsis    [Reads the command line options]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void
bdsReadOptions(
  int  argc,
  char ** argv,
  bdsOptions * option)
{
    int	i = 0;

    if (argc < 2) goto usage;

    for (i = 1; i < argc; i++) {
	if (argv[i][0] != '-' ) {
	    if (option->file == NULL) {
		option->file = util_strsav(argv[i]);
	    } else {
		goto usage;
	    }
	} else if (STRING_EQUAL(argv[i],"-v")) {
	    printf ("%s\n", BDS_VERSION);
	    exit(0);
	} else if (STRING_EQUAL(argv[i],"-verify")) {
	    option->verify = TRUE;
	} else if (STRING_EQUAL(argv[i],"-useloc")) {
	    option->useloc = TRUE;
	} else if (STRING_EQUAL(argv[i],"-useglb")) {
	    option->useglb = TRUE;
	} else if (STRING_EQUAL(argv[i],"-autodyn")) {
	    option->autodyn = TRUE;
	} else if (STRING_EQUAL(argv[i],"-rodelim")) {
	    option->rodelim = TRUE;
	} else if (STRING_EQUAL(argv[i],"-glbEliminate")) {
	    option->glbEliminate = TRUE;
	} else if (STRING_EQUAL(argv[i],"-cache")) {
	    i++;
	    option->cacheSize = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-maxmem")) {
	    i++;
	    option->maxMemory = 1048576 * (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-slots")) {
	    i++;
	    option->slots = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-time")) {
	    i++;
	    option->timeout = (long) atol(argv[i]) * 1000L;
	} else if (STRING_EQUAL(argv[i],"-order")) {
	    i++;
	    option->order = PI_PS_GIVEN;
	    option->orderfile = util_strsav(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-reordering")) {
	    i++;
	    if (STRING_EQUAL(argv[i],"none")) {
		option->reordering = CUDD_REORDER_NONE;
	    } else if (STRING_EQUAL(argv[i],"random")) {
		option->reordering = CUDD_REORDER_RANDOM;
	    } else if (STRING_EQUAL(argv[i],"bernard") ||
		STRING_EQUAL(argv[i],"pivot")) {
		option->reordering = CUDD_REORDER_RANDOM_PIVOT;
	    } else if (STRING_EQUAL(argv[i],"sifting")) {
		option->reordering = CUDD_REORDER_SIFT;
	    } else if (STRING_EQUAL(argv[i],"converge")) {
		option->reordering = CUDD_REORDER_SIFT_CONVERGE;
	    } else if (STRING_EQUAL(argv[i],"symm")) {
		option->reordering = CUDD_REORDER_SYMM_SIFT;
	    } else if (STRING_EQUAL(argv[i],"cosymm")) {
		option->reordering = CUDD_REORDER_SYMM_SIFT_CONV;
	    } else if (STRING_EQUAL(argv[i],"tree") ||
		STRING_EQUAL(argv[i],"group")) {
		option->reordering = CUDD_REORDER_GROUP_SIFT;
	    } else if (STRING_EQUAL(argv[i],"cotree") ||
		STRING_EQUAL(argv[i],"cogroup")) {
		option->reordering = CUDD_REORDER_GROUP_SIFT_CONV;
	    } else if (STRING_EQUAL(argv[i],"win2")) {
		option->reordering = CUDD_REORDER_WINDOW2;
	    } else if (STRING_EQUAL(argv[i],"win3")) {
		option->reordering = CUDD_REORDER_WINDOW3;
	    } else if (STRING_EQUAL(argv[i],"win4")) {
		option->reordering = CUDD_REORDER_WINDOW4;
	    } else if (STRING_EQUAL(argv[i],"win2conv")) {
		option->reordering = CUDD_REORDER_WINDOW2_CONV;
	    } else if (STRING_EQUAL(argv[i],"win3conv")) {
		option->reordering = CUDD_REORDER_WINDOW3_CONV;
	    } else if (STRING_EQUAL(argv[i],"win4conv")) {
		option->reordering = CUDD_REORDER_WINDOW4_CONV;
	    } else if (STRING_EQUAL(argv[i],"annealing")) {
		option->reordering = CUDD_REORDER_ANNEALING;
	    } else if (STRING_EQUAL(argv[i],"genetic")) {
		option->reordering = CUDD_REORDER_GENETIC;
	    } else if (STRING_EQUAL(argv[i],"linear")) {
		option->reordering = CUDD_REORDER_LINEAR;
	    } else if (STRING_EQUAL(argv[i],"linconv")) {
		option->reordering = CUDD_REORDER_LINEAR_CONVERGE;
	    } else if (STRING_EQUAL(argv[i],"exact")) {
		option->reordering = CUDD_REORDER_EXACT;
	    } else {
		goto usage;
	    }
	} else if (STRING_EQUAL(argv[i],"-growth")) {
	    i++;
	    option->maxGrowth = (int)atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-verb")) {
	    i++;
	    option->verb = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-miniMethod")) {
	    i++;
	    if (STRING_EQUAL(argv[i],"restrict")) {
		option->miniMethod = BDS_BDD_RESTRICT;
	    } else if (STRING_EQUAL(argv[i],"licompaction")) {
		option->miniMethod = BDS_BDD_LICOMPACTION;
	    } else {
		goto usage;
	    }
	} else if (STRING_EQUAL(argv[i],"-effort")) {
	    i++;
	    option->effort = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-limitSize")) {
	    i++;
	    option->limitSize = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-globalLimit")) {
	    i++;
	    option->globalLimit = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-ethred")) {
	    i++;
	    option->ethred = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-dumpbdd")) {
	    option->dumpbdd = TRUE;
	} else if (STRING_EQUAL(argv[i],"-dumpblif")) {
	    option->dumpblif = TRUE;
	} else if (STRING_EQUAL(argv[i],"-dumpftree")) {
	    option->dumpftree = TRUE;
	} else if (STRING_EQUAL(argv[i],"-dumpfdot")) {
	    option->dumpfdot = TRUE;
	} else if (STRING_EQUAL(argv[i],"-dumpnet")) {
	    option->dumpnet = TRUE;
	} else if (STRING_EQUAL(argv[i],"-cutptl")) {
	    option->cutptl = TRUE;
	} else if (STRING_EQUAL(argv[i],"-delta")) {
	    i++;
	    option->xdelta = (int) atoi(argv[i]);
	} else if (STRING_EQUAL(argv[i],"-dir")) {
	    i++;
	    option->dir = argv[i];
    } else if (STRING_EQUAL(argv[i],"-xhc")) {      /* x hardcore off JS */
            option->methods &= ~BDS_CTRL_XHC;
    } else if (STRING_EQUAL(argv[i],"-bhc")) {      /* branch hardcore off JS */
            option->methods &= ~BDS_CTRL_BHC;
    } else if (STRING_EQUAL(argv[i],"-sid")) {      /* simple dominators off JS */
            option->methods &= ~BDS_CTRL_SID;
    } else if (STRING_EQUAL(argv[i],"-ged")) {      /* generalized dominators off JS */
            option->methods &= ~BDS_CTRL_GED;
    } else if (STRING_EQUAL(argv[i],"-cft")) {      /* cofactor on top off JS */
            option->methods &= ~BDS_CTRL_CFT;
	} else {
	    goto usage;
	}
    }

    if (option->useloc && option->useglb) {
        printf("\nCommand line error: -useloc and -useglb can NOT be both specified !\n\n");
	exit(0);
    }
    if (option->file == NULL) {
        printf("\nPlease specify a circuit file !\n\n");
	exit(0);
    }

    (void) printf("# %s\n", BDS_VERSION);

    /* echo command line and arguments */
    (void) printf("#");
    for (i = 0; i < argc; i++) {
	(void) printf(" %s", argv[i]);
    }
    (void) printf("\n");
    (void) printf("# CUDD Version ");
    Cudd_PrintVersion(stdout);
    (void) fflush(stdout);
    printf("\n\n");
    BDS_Stats(BDS_INIT,NULL,argc,BDS_VERSION,argv,NULL,option);

    return;

usage:	/* convenient goto */
    if (i == 0) {
	(void) fprintf(stdout,"too few arguments\n");
    } else {
	(void) fprintf(stdout,"option: %s is not defined\n",argv[i]);
    }

    printf("Usage: bds [-option ] ckt.blif\n");
    printf("Options:\n");
    printf("\t-v\tPrint BDS version and exit.\n");
    printf("BDD control\n");
    printf("\t-order id\tThe initial variable order of the BDD.\n");
    printf("\t-rodelim\tNumber of BDD nodes after variable reordering.\n" );
    printf("\t\tDefault is TRUE.\n");
    printf("\t-limitSize n\tThe largest BDD size \n\t\ton which exact variable reordering can apply.\n");
    printf("\t\tDefault is set to 200 which is a fairly small BDD.\n");
    printf("\t-miniMethod method\tSelect the BDD minimization method.\n");
    printf("\t\tCurrently restrict and licompaction are supported. \n\t\tDefault is restrict.\n");
    printf("\t-cache n\tInitial cache size for CUDD. Default is 10K.\n");
    printf("\t-maxmem n\t Memory size limit for CUDD in megabytes. Default is 10.\n");
    printf("\t-reordering method\tSelect the BDD variable reordering in CUDD.\n"
           "\t\tCurrently supported: none, random, bernard, pivot,\n"
           "\t\tsifting, converge, symm, tree, cotree, group, \n"
           "\t\tcogroup, win2, win3, win4, win2conv, win3conv,\n"
           "\t\t win4conv, annealing, genetic, linear,liconv, and exact.\n"
           "\t\tDefault is set to none.\n");

    printf("Information output control\n");
    printf("\t-verb n\tverbosity level. There are verbosity levels 0, 1 and 4.\n");

    printf("Algorithm control\n");
    printf("\t-time n\tSpecify a timeout in seconds for preprocessing and decomposition phase.\n"
           "\t\tDisabled by default.\n");
    printf("\t-ethred n\tSpecify the eliminate gain threshold.\n\t\tDefault is set to 0.\n");
    printf("\t-delta n\tSpecify the disqualification threshold for X Hardcore.\n"
           "\t\tThe threshold value 0 forces decomposition monotonic in size.\n"
           "\t\tDefault is set to INT_MAX, which disables this feature.\n");
    printf("\t-useloc\tForce BDS to use local BDDs. Default is FALSE.\n");
    printf("\t-useloc\tForce BDS to use global BDDs. Default is FALSE.\n");
    printf("\t-sid\tDo not perform Simple Dominator (algebraic) decomposition.\n");
    printf("\t-ged\tDo not perform Generalized Dominators decomposition.\n");
    printf("\t-bhc\tDo not perform Branch Hardcore decomposition.\n");
    printf("\t-xhc\tDo not perform X Hardcore (XOR) decomposition.\n");
    printf("\t-cft\tDo not perform Cofactor on Top decomposition.\n");
    printf("\t-globalLimit n\tSet the total number of BDD nodes.\n"
           "\t\tDefault is set to 20,000.\n");
    printf("\t-glbEliminate\tEliminate all functional equivalent nodes.\n"
           "\t\tGlobal BDDs must be built first. Default FALSE.\n");
    printf("\t-verify\tPerfrom verification. Default is FALSE.\n");
    printf("Output control\n");
    printf("\t-dir dir-name\tForce BDS to use dir-name as the output directory.\n"
           "\t\tdir-name must not already exist.\n");
    printf("\t-dumpnet\tThe Boolean network after eliminate is dumped\n"
           "\t\tinto a file netdebug.dot in the current dir.\n"
           "\t\tDefault is set to FALSE.\n");
    printf("\t-dumpftree\tWrite the factoring tree in text (EQN format).\n"
           "\t\tDefault is set to TRUE.\n");
    printf("\t-dumpblif\tWrite the factoring tree in BLIF format.\n"
           "\t\tDefault is set to TRUE.\n");
    printf("\t-dumpfdot\tWrite the factoring tree in dot format.\n"
           "\t\tDefault is set to FALSE.\n");
    printf("\t-cutptl\tCut XORs, XNORs and MUX out from synthesized netlist.\n");
    exit(-1);

} /* end of bdsReadOptions */

/**Function********************************************************************

  Synopsis    [Frees the option structure and its appendages.]

  Description []

  SideEffects [None]

  SeeAlso     []

*****************************************************************************/
static void
freeOption(
  bdsOptions * option)
{
    if (option->file != NULL) FREE(option->file);
    if (option->orderfile != NULL) FREE(option->orderfile);
    FREE(option);

} /* end of freeOption */

/**Function********************************************************************

  Synopsis    [Starts the CUDD manager with the desired options.]

  Description []

  SideEffects [None]

  SeeAlso     []

*****************************************************************************/
EXTERN
DdManager *
startCudd(
  bdsOptions * option,
  int  nvars)
{
    DdManager *dd;
    int result;

    dd = Cudd_Init(0, 0, option->slots, option->cacheSize, option->maxMemory);
    if (dd == NULL) return(NULL);

    Cudd_SetGroupcheck(dd,CUDD_GROUP_CHECK7);

/* Dynamic variable reordering is controlled by program. -cyang 5/16/99
    if (option->autodyn) {
        Cudd_AutodynEnable(dd,CUDD_REORDER_SIFT);
    }
*/

    dd->nextDyn = DD_FIRST_REORDER;
    dd->countDead = FALSE;
    dd->maxGrowth = 1.0 + ((float) option->maxGrowth / 100.0);
    dd->recomb = DD_DEFAULT_RECOMB;
    dd->arcviolation = 10;
    dd->symmviolation = 10;
    dd->populationSize = 0;
    dd->numberXovers = 0;

#ifndef DD_STATS
    result = Cudd_EnableReorderingReporting(dd);
    if (result == 0) {
        (void) fprintf(stderr,
                       "Error reported by Cudd_EnableReorderingReporting\n");
        Cudd_Quit(dd);
        return(NULL);
    }
#endif

    return(dd);

} /* end of startCudd */


